package main

import (
	"code.mrmelon54.com/melon/summer/cmd/azalea/config"
	domainChecker "code.mrmelon54.com/melon/summer/cmd/azalea/domain-checker"
	"code.mrmelon54.com/melon/summer/cmd/azalea/favicons"
	quickDb "code.mrmelon54.com/melon/summer/cmd/azalea/quick-db"
	"code.mrmelon54.com/melon/summer/cmd/azalea/routing"
	"code.mrmelon54.com/melon/summer/cmd/azalea/servers"
	"code.mrmelon54.com/melon/summer/pkg/api"
	"code.mrmelon54.com/melon/summer/pkg/cli"
	"code.mrmelon54.com/melon/summer/pkg/proxy"
	"code.mrmelon54.com/melon/summer/pkg/tables/certificate"
	"code.mrmelon54.com/melon/summer/pkg/tables/renewal"
	"code.mrmelon54.com/melon/summer/pkg/tables/web"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"errors"
	"github.com/kjk/common/filerotate"
	"github.com/mrmelon54/mjwt"
	"io"
	"net/http"
	"net/http/httputil"
	"os"
	"strconv"
	"strings"
)

var usedTables = []any{
	// web
	&web.Domain{},
	&web.ApiRoute{},
	&web.ApiDomain{},
	&web.HttpRoute{},
	&web.HttpService{},
	&web.HttpRedirect{},
	&web.HttpRouteHeader{},
	&web.HttpServiceHeader{},
	&web.HttpDefaultHost{},
	&web.HttpFavicon{},

	// renewal
	&renewal.AcmeContent{},

	// certificate
	&certificate.Certificate{},
	&certificate.CertificateData{},
	&certificate.CertificateDomain{},
}

func main() {
	cli.RunWithConfig[config.AzaleaConfig]("azalea", func(t config.AzaleaConfig) string { return t.Database }, new(Azalea))
}

type Azalea struct {
	conf         config.AzaleaConfig
	runner       *cli.Runner[config.AzaleaConfig]
	reverseProxy *httputil.ReverseProxy
	httpServer   *http.Server
	httpsServer  *http.Server
	apiServer    *http.Server
	verify       mjwt.Provider
	accessLog    *filerotate.File
}

func (azalea *Azalea) Init(runner *cli.Runner[config.AzaleaConfig]) {
	azalea.runner = runner
	azalea.conf = runner.Config
	utils.Check("[Azalea.Init()] Failed to sync tables", runner.Database.Sync(usedTables...))
	quickDb.QuickDb(runner.Database)

	fourEighteenErrorFile, err := os.Open(azalea.conf.Code418)
	utils.Check("[Azalea.Init()] 418 Error File", err)
	fourEighteenErrorBytes, err := io.ReadAll(fourEighteenErrorFile)
	utils.Check("[Azalea.Init()] 418 Error File", err)

	fourEighteenErrorFunc := func(rw http.ResponseWriter, req *http.Request) {
		rw.WriteHeader(http.StatusTeapot)
		_, _ = rw.Write(fourEighteenErrorBytes)
	}

	daily, err := filerotate.NewDaily(azalea.conf.Logs.AccessLogDir, nil)
	utils.Check("[Azalea.Init()] Failed to create new daily access log rotation", err)
	azalea.accessLog = daily

	addrPort, err := splitPortCombo(azalea.conf.Listen.Https)
	utils.Check("[Azalea.Init()] Failed to parse HTTPs port", err)
	azalea.verify, err = mjwt.NewMJwtVerifierFromFile(azalea.conf.Auth.Public)
	utils.Check("[Azalea.Init()] Failed to load verifier public key", err)

	azalea.reverseProxy = proxy.CreateHybridReverseProxy()
	tableRouter, err := routing.New(runner.Database, azalea.reverseProxy, azalea.accessLog, fourEighteenErrorFunc)
	domainCheck := domainChecker.New(runner.Database)
	faviconMap := favicons.New(runner.Database, azalea.conf.Inkscape)
	compileTargets := api.MultiCompilable{tableRouter, domainCheck, faviconMap}
	utils.Check("[Azalea.Init()] Failed to load table router:", err)

	azalea.httpServer = servers.NewHttpServer(azalea.conf.Listen.Http, addrPort, azalea.runner.Database, domainCheck)
	azalea.httpsServer = servers.NewHttpsServer(azalea.conf.Listen.Https, azalea.runner.Database, tableRouter, domainCheck, faviconMap, azalea.conf, fourEighteenErrorFunc)
	azalea.apiServer = servers.NewApiServer(azalea.conf.Listen.Api, azalea.runner.Database, azalea.verify, compileTargets)
}

func (azalea *Azalea) Destroy() {
	_ = azalea.httpServer.Close()
	_ = azalea.httpsServer.Close()
	_ = azalea.apiServer.Close()
}

func splitPortCombo(a string) (uint16, error) {
	i := strings.LastIndexByte(a, ':')
	if i == -1 {
		return 0, errors.New("not an ip:port")
	}
	port := a[i+1:]
	parseInt, err := strconv.ParseUint(port, 10, 16)
	if err != nil {
		return 0, err
	}
	return uint16(parseInt), err
}
