//go:build DEBUG || TEST

package quick_db

import (
	"code.mrmelon54.com/melon/certgen"
	"code.mrmelon54.com/melon/summer/pkg/tables/certificate"
	"code.mrmelon54.com/melon/summer/pkg/tables/web"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"crypto/x509/pkix"
	"fmt"
	"log"
	"math/big"
	"math/rand"
	"net"
	"time"
	"xorm.io/xorm"
)

func QuickDb(db *xorm.Engine) {
	quickCert(db)
	quickDomain(db)
	quickApi(db)
}

func quickCert(db *xorm.Engine) {
	log.Println("[quickCert()] Generating development certificate for testing azalea")

	count, err := db.Count(&certificate.Certificate{})
	utils.Check("[quickCert()] Failed to count certificates:", err)
	if count != 0 {
		return
	}

	n := time.Now()
	caTls, err := certgen.MakeCaTls(pkix.Name{CommonName: "ca.summer.test"}, new(big.Int).SetUint64(rand.Uint64()))
	utils.Check("[quickCert()] Failed to make CA TLS:", err)
	serverTls, err := certgen.MakeServerTls(caTls, pkix.Name{CommonName: "summer.test"}, new(big.Int).SetUint64(rand.Uint64()), []string{"summer.test", "*.summer.test"}, []net.IP{})
	utils.Check("[quickCert()] Failed to make server TLS:", err)
	cert := serverTls.GetTlsLeaf()
	leaf := certgen.TlsLeaf(&cert)

	c := certificate.Certificate{
		Owner:       1,
		LetsEncrypt: 1,
		Namesilo:    1,
		AutoRenew:   utils.PBool(false),
		Active:      utils.PBool(true),
		Renewing:    utils.PBool(false),
		RenewFailed: utils.PBool(false),
	}
	_, err = db.Insert(&c)
	utils.Check("[quickCert()] Failed to insert certificate:", err)
	_, err = db.Insert(&certificate.CertificateData{
		MetaId:           c.Id,
		PreviousCert:     0,
		CertificateBytes: serverTls.GetCertPem(),
		KeyBytes:         serverTls.GetKeyPem(),
		NotAfter:         leaf.NotAfter,
		CreatedAt:        n,
		UpdatedAt:        n,
		Ready:            utils.PBool(true),
	})
	utils.Check("[quickCert()] Failed to insert certificate data:", err)
	_, err = db.Insert(&certificate.CertificateDomain{
		DomainId: 0,
		CertId:   c.Id,
		Domain:   "summer.test",
		Wildcard: utils.PBool(false),
	}, &certificate.CertificateDomain{
		DomainId: 0,
		CertId:   c.Id,
		Domain:   "summer.test",
		Wildcard: utils.PBool(true),
	})
	utils.Check("[quickCert()] Failed to insert certificate domains:", err)

	log.Println("[quickCert()] Finished generating development certificate")
}

func quickDomain(db *xorm.Engine) {
	log.Println("[quickDomain()] Generating development domain for testing azalea")

	count, err := db.Count(&web.Domain{})
	utils.Check("[quickDomain()] Failed to count domains:", err)
	if count != 0 {
		return
	}

	_, err = db.Insert(&web.Domain{Domain: "summer.test"})
	utils.Check("[quickDomain()] Failed to insert domain:", err)
}

func quickApi(db *xorm.Engine) {
	log.Println("[quickApi()] Generating development API for testing azalea")

	count, err := db.Count(&web.ApiDomain{})
	utils.Check("[quickApi()] Failed to count API domains:", err)
	if count == 0 {
		_, err = db.Insert(&web.ApiDomain{
			Domain:  "api.summer.test",
			Enabled: utils.PBool(true),
		})
		utils.Check("[quickApi()] Failed to insert API domain:", err)
	}

	apiDomain := web.ApiDomain{Domain: "api.summer.test"}
	get, err := db.Get(&apiDomain)
	utils.Check("[quickApi()] Failed to get API domain:", err)
	if !get {
		utils.Check("[quickApi()] Failed to get API domain:", fmt.Errorf("no such record"))
	}

	count, err = db.Count(&web.ApiRoute{})
	utils.Check("[quickApi()] Failed to count API routes:", err)
	if count == 0 {
		_, err = db.Insert(&web.ApiRoute{
			DomainId:   apiDomain.Id,
			Version:    1,
			Route:      "azalea",
			Dst:        "azalea:7070",
			SecureMode: utils.PBool(false),
			IgnoreCert: utils.PBool(false),
			Enabled:    utils.PBool(true),
		})
		_, err = db.Insert(&web.ApiRoute{
			DomainId:   apiDomain.Id,
			Version:    1,
			Route:      "buttercup",
			Dst:        "buttercup:7070",
			SecureMode: utils.PBool(false),
			IgnoreCert: utils.PBool(false),
			Enabled:    utils.PBool(true),
		})
		_, err = db.Insert(&web.ApiRoute{
			DomainId:   apiDomain.Id,
			Version:    1,
			Route:      "marigold",
			Dst:        "marigold:7070",
			SecureMode: utils.PBool(false),
			IgnoreCert: utils.PBool(false),
			Enabled:    utils.PBool(true),
		})
		_, err = db.Insert(&web.ApiRoute{
			DomainId:   apiDomain.Id,
			Version:    1,
			Route:      "rose",
			Dst:        "rose:7070",
			SecureMode: utils.PBool(false),
			IgnoreCert: utils.PBool(false),
			Enabled:    utils.PBool(true),
		})
	}
}
