package routing

import (
	"code.mrmelon54.com/melon/summer/pkg/tables/web"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"errors"
	"fmt"
	"github.com/gorilla/mux"
	"github.com/kjk/common/filerotate"
	"log"
	"net/http"
	"net/http/httputil"
	"sync"
	"xorm.io/xorm"
)

var ErrAlreadyCompiling = errors.New("router: already compiling")
var theIncrementor uint64

type TableRouter struct {
	router       *mux.Router
	db           *xorm.Engine
	proxy        *httputil.ReverseProxy
	rLock        *sync.RWMutex // lock for router
	cLock        *sync.Mutex   // lock for internalCompile
	accessLog    *log.Logger
	fourEighteen func(rw http.ResponseWriter, req *http.Request)
}

func New(db *xorm.Engine, proxy *httputil.ReverseProxy, accessLog *filerotate.File, fourEighteenError func(rw http.ResponseWriter, req *http.Request)) (*TableRouter, error) {
	r := &TableRouter{
		router:       mux.NewRouter(),
		db:           db,
		proxy:        proxy,
		rLock:        &sync.RWMutex{},
		cLock:        &sync.Mutex{},
		accessLog:    log.New(accessLog, "", log.LstdFlags),
		fourEighteen: fourEighteenError,
	}
	// run internalCompile inline to generate the initial router
	err := r.internalCompile(r.router)
	return r, err
}

func (t *TableRouter) ServeHTTP(rw http.ResponseWriter, req *http.Request) {
	// checks the router is not locked by Compile while swapping router
	t.rLock.RLock()
	theIncrementor++
	z := theIncrementor
	r := t.router
	t.rLock.RUnlock()
	t.accessLog.Printf("[TableRouter %d] Serving '%s' - '%s'\n", z, req.Host, req.URL.String())
	r.ServeHTTP(rw, req)
	t.accessLog.Printf("[TableRouter %d] Served request\n", z)
}

func (t *TableRouter) Compile() {
	// start the goroutine and leave
	go func() {
		// create a new router and call internalCompile
		router := mux.NewRouter()
		err := t.internalCompile(router)
		if err != nil {
			// log compile errors
			log.Printf("[TableRouter] Compile failed: %s\n", err)
			return
		}
		// lock while replacing the router
		t.rLock.Lock()
		t.router = router
		t.rLock.Unlock()
	}()
}

func (t *TableRouter) internalCompile(router *mux.Router) error {
	if !t.cLock.TryLock() {
		return ErrAlreadyCompiling
	}
	defer t.cLock.Unlock()

	router.NotFoundHandler = http.HandlerFunc(t.fourEighteen)

	log.Println("[TableRouter] Pulling from database")
	var (
		apiDomain          []web.ApiDomain
		apiRoute           []web.ApiRoute
		httpDefaultHost    []web.HttpDefaultHost
		httpRoutes         []web.HttpRoute
		httpServices       []web.HttpService
		httpRedirects      []web.HttpRedirect
		httpRouteHeaders   []web.HttpRouteHeader
		httpServiceHeaders []web.HttpServiceHeader
	)
	if err := t.db.Find(&apiDomain, web.ApiDomain{Enabled: utils.PBool(true)}); err != nil {
		return err
	}
	if err := t.db.Find(&apiRoute, web.ApiRoute{Enabled: utils.PBool(true)}); err != nil {
		return err
	}
	if err := t.db.Find(&httpDefaultHost); err != nil {
		return err
	}
	if err := t.db.Find(&httpRoutes, web.HttpRoute{Enabled: utils.PBool(true)}); err != nil {
		return err
	}
	if err := t.db.Find(&httpServices, web.HttpService{Enabled: utils.PBool(true)}); err != nil {
		return err
	}
	if err := t.db.Find(&httpRedirects, web.HttpRedirect{Enabled: utils.PBool(true)}); err != nil {
		return err
	}
	if err := t.db.Find(&httpRouteHeaders); err != nil {
		return err
	}
	if err := t.db.Find(&httpServiceHeaders); err != nil {
		return err
	}

	log.Println("[TableRouter] Mapping default hosts")
	mapDefaultHost := make(map[string]web.HttpDefaultHost)
	for _, i := range httpDefaultHost {
		if _, ok := mapDefaultHost[i.Key]; ok {
			return fmt.Errorf("DefaultHost[%s] already exists", i.Key)
		}
		mapDefaultHost[i.Key] = i
	}

	log.Println("[TableRouter] Mapping api routes to domains")
	mapApiRoute := make(map[uint64][]web.ApiRoute)
	for _, i := range apiRoute {
		if mapApiRoute[i.DomainId] == nil {
			mapApiRoute[i.DomainId] = make([]web.ApiRoute, 0)
		}
		mapApiRoute[i.DomainId] = append(mapApiRoute[i.DomainId], i)
	}

	log.Println("[TableRouter] Mapping modified headers")
	mapRouteHeaders := make(map[uint64][]ServeHeaderItem)
	mapServiceHeaders := make(map[uint64][]ServeHeaderItem)
	for _, i := range httpRouteHeaders {
		if mapRouteHeaders[i.RouteId] == nil {
			mapRouteHeaders[i.RouteId] = make([]ServeHeaderItem, 0)
		}
		mapRouteHeaders[i.RouteId] = append(mapRouteHeaders[i.RouteId], i)
	}
	for _, i := range httpServiceHeaders {
		if mapServiceHeaders[i.ServiceId] == nil {
			mapServiceHeaders[i.ServiceId] = make([]ServeHeaderItem, 0)
		}
		mapServiceHeaders[i.ServiceId] = append(mapServiceHeaders[i.ServiceId], i)
	}

	log.Println("[TableRouter] Adding api routes")
	for _, i := range apiDomain {
		routes := mapApiRoute[i.Id]
		if routes == nil {
			continue
		}
		subR := router.Host(i.Domain).Subrouter()
		for _, j := range routes {
			pre := fmt.Sprintf("/v%d/%s", j.Version, j.Route)
			hostD, portD, pathD, _, refD, err := utils.ParseSymbolicHost(j.Dst)
			if err != nil {
				return fmt.Errorf("%w: '%s'", err, j.Dst)
			}

			if refD {
				if err := resolveRef(mapDefaultHost, &hostD, &portD, &pathD); err != nil {
					return err
				}
			}

			subR.PathPrefix(pre).Handler(ServeRoute{Host: hostD, Port: portD, Path: pathD, PathPre: pre, SecureMode: j.IsSecureMode(), IgnoreCert: j.IsIgnoreCert(), Proxy: t.proxy, Cors: true})
		}
	}

	log.Println("[TableRouter] Adding redirects")
	for _, i := range httpRedirects {
		host, _, path, flags, _, err := utils.ParseSymbolicHost(i.Src)
		if err != nil {
			return fmt.Errorf("%w: '%s'", err, i.Src)
		}

		hostD, portD, pathD, flagsD, _, err := utils.ParseSymbolicHost(i.Target)

		z := router.Host(host)
		if path != "" {
			if utils.Contains[string](flags, "prefix") {
				z = z.PathPrefix(path)
			} else {
				z = z.Path(path)
			}
		} else {
			z = z.PathPrefix("/")
		}
		if portD != 0 {
			hostD = fmt.Sprintf("%s:%d", hostD, portD)
		}

		z.Handler(ServeRedirect{Target: hostD, Path: pathD, Prefix: path, Flags: flagsD})
	}

	log.Println("[TableRouter] Adding routes")
	for _, i := range httpRoutes {
		hostS, _, pathS, _, _, err := utils.ParseSymbolicHost(i.Src)
		if err != nil {
			return fmt.Errorf("%w: '%s'", err, i.Src)
		}
		hostD, portD, pathD, _, refD, err := utils.ParseSymbolicHost(i.Dst)
		if err != nil {
			return fmt.Errorf("%w: '%s'", err, i.Dst)
		}

		if refD {
			if err := resolveRef(mapDefaultHost, &hostD, &portD, &pathD); err != nil {
				return err
			}
		}

		z := router.Host(hostS)
		if pathS != "" {
			z = z.Path(pathS)
		}
		z.Handler(ServeRoute{Host: hostD, Port: portD, Path: pathD, Headers: mapRouteHeaders[i.Id], AbsPath: i.IsAbsPath(), SecureMode: i.IsSecureMode(), ForwardHost: i.IsForwardHost(), IgnoreCert: i.IsIgnoreCert(), Proxy: t.proxy})
	}

	log.Println("[TableRouter] Adding services")
	for _, i := range httpServices {
		hostD, portD, pathD, _, refD, err := utils.ParseSymbolicHost(i.Dst)
		if err != nil {
			return err
		}

		if refD {
			if err := resolveRef(mapDefaultHost, &hostD, &portD, &pathD); err != nil {
				return err
			}
		}
		router.Host(i.Host).Handler(ServeRoute{Host: hostD, Port: portD, Path: pathD, Headers: mapServiceHeaders[i.Id], AbsPath: i.IsAbsPath(), SecureMode: i.IsSecureMode(), ForwardHost: i.IsForwardHost(), IgnoreCert: i.IsIgnoreCert(), Proxy: t.proxy})
	}

	// well I guess we are done
	log.Println("[TableRouter] Finished compiling, updated routes will be available shortly")
	return nil
}

func resolveRef(mapDefaultHost map[string]web.HttpDefaultHost, host *string, port *uint16, path *string) error {
	if y, ok := mapDefaultHost[(*host)[1:]]; ok {
		a1, a2, a3, _, _, err := utils.ParseSymbolicHost(y.Dst)
		if err != nil {
			return err
		}
		*host = a1
		if *port == 0 {
			*port = a2
		}
		if *path == "" {
			*path = a3
		}
	}
	return nil
}
