package routing

import (
	"bytes"
	"code.mrmelon54.com/melon/summer/pkg/proxy"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"fmt"
	"github.com/gorilla/handlers"
	"io"
	"log"
	"net/http"
	"net/url"
	"path"
	"strings"
)

var serveApiCors = handlers.CORS(
	handlers.AllowCredentials(),
	handlers.AllowedOrigins([]string{"*"}),
	handlers.AllowedHeaders([]string{"Content-Type", "Authorization"}),
	handlers.AllowedMethods([]string{
		http.MethodGet,
		http.MethodHead,
		http.MethodPost,
		http.MethodPut,
		http.MethodPatch,
		http.MethodDelete,
		http.MethodConnect,
		http.MethodOptions,
		http.MethodTrace,
	}),
)

type ServeRoute struct {
	Host        string
	Port        uint16
	Path        string
	PathPre     string // prefix from http route specific call
	Headers     []ServeHeaderItem
	Proxy       http.Handler
	AbsPath     bool
	SecureMode  bool
	ForwardHost bool
	IgnoreCert  bool
	Cors        bool
}

func (s ServeRoute) IsIgnoreCert() bool { return s.IgnoreCert }

func (s ServeRoute) UpdateHeaders(header http.Header) {
	for _, i := range s.Headers {
		header.Set(i.HeaderKey(), i.HeaderValue())
	}
}

func (s ServeRoute) ServeHTTP(rw http.ResponseWriter, req *http.Request) {
	if s.Cors {
		serveApiCors(http.HandlerFunc(s.internalServeHTTP)).ServeHTTP(rw, req)
	} else {
		s.internalServeHTTP(rw, req)
	}
}

func (s ServeRoute) internalServeHTTP(rw http.ResponseWriter, req *http.Request) {
	scheme := "http"
	if s.SecureMode {
		scheme = "https"
		if s.Port == 0 {
			s.Port = 443
		}
	} else {
		if s.Port == 0 {
			s.Port = 80
		}
	}
	var p1 string
	if s.AbsPath {
		p1 = path.Join(s.Path)
	} else {
		if !strings.HasPrefix(req.URL.Path, s.PathPre) {
			utils.RespondHttpStatus(rw, http.StatusExpectationFailed)
			return
		}
		p1 = path.Join(s.Path, req.URL.Path[len(s.PathPre):])

		// add suffix "/" to prevent redirect issues
		// the path.Join() above calls path.Clean() and strips the trailing slash
		// so this just adds the trailing slash back in
		if strings.HasSuffix(req.URL.Path, "/") && !strings.HasSuffix(p1, "/") {
			p1 += "/"
		}
	}

	// If path is empty then correct to /
	if p1 == "" {
		p1 = "/"
	}

	buf := new(bytes.Buffer)
	if req.Body != nil {
		_, _ = io.Copy(buf, req.Body)
	}

	u := url.URL{
		Scheme:   scheme,
		Host:     fmt.Sprintf("%s:%d", s.Host, s.Port),
		Path:     p1,
		RawQuery: req.URL.RawQuery,
	}
	req2, err := http.NewRequest(req.Method, u.String(), buf)
	if err != nil {
		log.Printf("[ServeRoute::ServerHTTP()] Error generating new request: %s\n", err)
		utils.RespondHttpStatus(rw, http.StatusBadGateway)
		return
	}
	for k, v := range req.Header {
		if k == "Host" {
			continue
		}
		req2.Header[k] = v
	}
	if s.ForwardHost {
		req2.Host = req.Host
	}
	s.Proxy.ServeHTTP(rw, proxy.SetReverseProxyHost(req2, s))
}
