package routing

import (
	"bytes"
	"github.com/stretchr/testify/assert"
	httpTest "github.com/stretchr/testify/http"
	"net/http"
	"testing"
)

var routeLoopProxy = http.HandlerFunc(func(rw http.ResponseWriter, req *http.Request) {
	rw.Header().Set("Location", req.URL.String())
	rw.Header().Set("Host", req.Host)
})

func TestServeRoute_Location(t *testing.T) {
	// Minimal test
	routeLocationCheck(t, "http://localhost:8080/", "", "https://a.example.com", ServeRoute{
		Host: "localhost",
		Port: 8080,
	})
	routeLocationCheck(t, "http://localhost:80/goodbye/world", "", "https://a.example.com/world", ServeRoute{
		Host: "localhost",
		Port: 0,
		Path: "/goodbye",
	})

	// ForwardHost test
	routeLocationCheck(t, "http://localhost:80/", "a.example.com", "https://a.example.com", ServeRoute{
		Host:        "localhost",
		Port:        0,
		ForwardHost: true,
	})
	routeLocationCheck(t, "http://localhost:8080/goodbye/world", "a.example.com", "https://a.example.com/world", ServeRoute{
		Host:        "localhost",
		Port:        8080,
		Path:        "/goodbye",
		ForwardHost: true,
	})

	// SecureMode test
	routeLocationCheck(t, "https://localhost:443/", "a.example.com", "https://a.example.com", ServeRoute{
		Host:        "localhost",
		Port:        0,
		ForwardHost: true,
		SecureMode:  true,
	})
	routeLocationCheck(t, "https://localhost:8443/goodbye/world", "a.example.com", "https://a.example.com/world", ServeRoute{
		Host:        "localhost",
		Port:        8443,
		Path:        "/goodbye",
		ForwardHost: true,
		SecureMode:  true,
	})

	// AbsPath test
	routeLocationCheck(t, "http://localhost:80/", "a.example.com", "https://a.example.com", ServeRoute{
		Host:        "localhost",
		Port:        0,
		ForwardHost: true,
		AbsPath:     true,
	})
	routeLocationCheck(t, "http://localhost:8080/goodbye", "a.example.com", "https://a.example.com/world", ServeRoute{
		Host:        "localhost",
		Port:        8080,
		Path:        "/goodbye",
		ForwardHost: true,
		AbsPath:     true,
	})

	// TrailingSlash test
	routeLocationCheck(t, "http://localhost:80/", "a.example.com", "https://a.example.com/", ServeRoute{
		Host:        "localhost",
		Port:        0,
		ForwardHost: true,
	})
	routeLocationCheck(t, "http://localhost:8080/goodbye/world/", "a.example.com", "https://a.example.com/world/", ServeRoute{
		Host:        "localhost",
		Port:        8080,
		Path:        "/goodbye",
		ForwardHost: true,
	})

	// Query test
	routeLocationCheck(t, "http://localhost:80/?raw=true", "a.example.com", "https://a.example.com/?raw=true", ServeRoute{
		Host:        "localhost",
		Port:        0,
		ForwardHost: true,
	})
	routeLocationCheck(t, "http://localhost:8080/goodbye/world/?raw=true", "a.example.com", "https://a.example.com/world/?raw=true", ServeRoute{
		Host:        "localhost",
		Port:        8080,
		Path:        "/goodbye",
		ForwardHost: true,
	})
}

func routeLocationCheck(t *testing.T, expected, host, input string, serve ServeRoute) {
	serve.Proxy = routeLoopProxy
	rw := &httpTest.TestResponseWriter{}
	req, err := http.NewRequest(http.MethodGet, input, new(bytes.Buffer))
	assert.NoError(t, err)
	serve.ServeHTTP(rw, req)
	assert.Equal(t, expected, rw.Header().Get("Location"))
}
