package servers

import (
	"code.mrmelon54.com/melon/summer/cmd/azalea/config"
	domainChecker "code.mrmelon54.com/melon/summer/cmd/azalea/domain-checker"
	"code.mrmelon54.com/melon/summer/cmd/azalea/favicons"
	"code.mrmelon54.com/melon/summer/cmd/azalea/routing"
	"code.mrmelon54.com/melon/summer/pkg/tables/certificate"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"crypto/tls"
	"fmt"
	"github.com/gorilla/mux"
	"github.com/mrmelon54/favicon"
	"github.com/sethvargo/go-limiter/httplimit"
	"github.com/sethvargo/go-limiter/memorystore"
	"log"
	"net/http"
	"time"
	"xorm.io/xorm"
)

func NewHttpsServer(listen string, db *xorm.Engine, tableRouter *routing.TableRouter, domainCheck *domainChecker.DomainChecker, faviconMap *favicons.Favicons, conf config.AzaleaConfig, fourEighteenError func(rw http.ResponseWriter, req *http.Request)) *http.Server {
	router := mux.NewRouter()
	setupHttpsRouter(router, tableRouter, domainCheck, faviconMap, conf, fourEighteenError)

	certCache := utils.NewCertCache()

	s := &http.Server{
		Addr:    listen,
		Handler: router,
		TLSConfig: &tls.Config{
			GetCertificate: func(info *tls.ClientHelloInfo) (*tls.Certificate, error) {
				if !domainCheck.IsValid(info.ServerName) {
					return nil, fmt.Errorf("invalid hostname used: '%s'", info.ServerName)
				}

				if certGrab, ok := certCache.Get(info.ServerName); ok {
					return certGrab, nil
				}

				// Just grab, save and output the certificate now
				cert, err := certificate.GetCertForDomain(db, info.ServerName)
				if err != nil {
					return nil, err
				}
				certLeaf, err := cert.GetCertificate()
				if err != nil {
					return nil, err
				}
				certCache.Put(info.ServerName, certLeaf)
				return certLeaf, nil
			},
		},
		ReadTimeout:       150 * time.Second,
		ReadHeaderTimeout: 150 * time.Second,
		WriteTimeout:      150 * time.Second,
		IdleTimeout:       150 * time.Second,
		MaxHeaderBytes:    4096000,
	}
	s.SetKeepAlivesEnabled(false)
	log.Printf("[HTTPS] Starting HTTPS server on: '%s'\n", s.Addr)
	go utils.RunBackgroundHttps("HTTPS", s)
	return s
}

func setupHttpsRouter(router *mux.Router, tableRouter *routing.TableRouter, domainCheck *domainChecker.DomainChecker, faviconMap *favicons.Favicons, conf config.AzaleaConfig, fourEighteenError func(rw http.ResponseWriter, req *http.Request)) {
	faviconColor := favicon.NewColor()

	router.Use(setupRateLimiter(conf.RateLimit))
	router.HandleFunc("/favicon.{ext:(?:ico|png|svg)}", utils.CachedPage(func(rw http.ResponseWriter, req *http.Request) {
		if !domainCheck.IsValid(req.Host) {
			fourEighteenError(rw, req)
			return
		}

		domain, ok := utils.GetDomainWithoutPort(req.Host)
		if !ok {
			fourEighteenError(rw, req)
			return
		}

		vars := mux.Vars(req)
		ext := vars["ext"]
		h := rw.Header()

		var favSvg produceImg
		if faviconList, ok := faviconMap.GetIcons(domain); ok {
			favSvg = faviconList
		} else {
			favSvg = favicon.NewSvg(domain, faviconColor)
		}

		var b []byte
		var err error

		switch ext {
		case "svg":
			h.Set("Content-Type", "image/svg+xml")
			b, err = favSvg.ProduceSvg()
		case "png":
			h.Set("Content-Type", "image/png")
			b, err = favSvg.ProducePng()
		case "ico":
			h.Set("Content-Type", "image/x-icon")
			b, err = favSvg.ProduceIco()
		}

		if err != nil {
			utils.RespondHttpStatus(rw, http.StatusInternalServerError)
			return
		}

		rw.WriteHeader(http.StatusOK)
		_, _ = rw.Write(b)
	}))
	router.NotFoundHandler = http.HandlerFunc(func(rw http.ResponseWriter, req *http.Request) {
		if !domainCheck.IsValid(req.Host) {
			fourEighteenError(rw, req)
			return
		}

		tableRouter.ServeHTTP(rw, req)
	})
}

func setupRateLimiter(rateLimit uint64) mux.MiddlewareFunc {
	store, err := memorystore.New(&memorystore.Config{
		Tokens:   rateLimit,
		Interval: time.Minute,
	})
	if err != nil {
		log.Fatalln(err)
	}
	middleware, err := httplimit.NewMiddleware(store, httplimit.IPKeyFunc())
	if err != nil {
		log.Fatalln(err)
	}
	return middleware.Handle
}

type produceImg interface {
	ProduceIco() ([]byte, error)
	ProducePng() ([]byte, error)
	ProduceSvg() ([]byte, error)
}
