package api

import (
	"code.mrmelon54.com/melon/summer/pkg/api/crud"
	"code.mrmelon54.com/melon/summer/pkg/renewal"
	"code.mrmelon54.com/melon/summer/pkg/tables/certificate"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"strings"
	"xorm.io/xorm"
)

type CertDbItem struct {
	Id          uint64   `json:"id"`
	Owner       uint64   `json:"owner"`
	AutoRenew   bool     `json:"autoRenew"`
	Active      bool     `json:"active"`
	Renewing    bool     `json:"renewing"`
	RenewFailed bool     `json:"renewFailed"`
	Domains     []string `json:"domains"`
}

func NewCertDbProvider(db *xorm.Engine, renewal *renewal.Renewal) crud.CheckedProvider[CertDbItem] {
	return &CertDbProvider{db, renewal}
}

type CertDbProvider struct {
	db      *xorm.Engine
	renewal *renewal.Renewal
}

func (c CertDbProvider) Find() ([]CertDbItem, error) {
	rows := make([]CertDbItem, 0)
	err := c.db.Iterate(&certificate.Certificate{}, func(i int, bean any) error {
		a := bean.(*certificate.Certificate)
		var b []certificate.CertificateDomain
		err := c.db.Where("cert_id = ?", a.Id).Find(&b)
		if err != nil {
			b = []certificate.CertificateDomain{}
		}
		d := make([]string, len(b))
		for i := range b {
			if utils.SBool(b[i].Wildcard) {
				d[i] = "*." + b[i].Domain
			} else {
				d[i] = b[i].Domain
			}
		}
		rows = append(rows, CertDbItem{
			Id:          a.Id,
			Owner:       a.Owner,
			AutoRenew:   utils.SBool(a.AutoRenew),
			Active:      utils.SBool(a.Active),
			Renewing:    utils.SBool(a.Renewing),
			RenewFailed: utils.SBool(a.RenewFailed),
			Domains:     d,
		})
		return nil
	})
	return rows, err
}

func (c CertDbProvider) Add(item CertDbItem) (CertDbItem, error) {
	// TODO: add full owner, lets encrypt and namesilo support
	a := certificate.Certificate{
		Id:          0,
		Owner:       0,
		LetsEncrypt: 0,
		Namesilo:    0,
		AutoRenew:   utils.PBool(item.AutoRenew),
		Active:      utils.PBool(true),
		Renewing:    utils.PBool(false),
		RenewFailed: utils.PBool(false),
	}
	d := item.Domains
	_, err := c.db.Insert(&a)
	if err != nil {
		return CertDbItem{}, err
	}
	for _, z := range d {
		isWild := strings.HasPrefix(z, "*.")
		if isWild {
			z = "*." + z
		}
		_, err = c.db.Insert(&certificate.CertificateDomain{
			DomainId: 0,
			CertId:   a.Id,
			Domain:   z,
			Wildcard: utils.PBool(isWild),
		})
		if err != nil {
			return CertDbItem{}, err
		}
	}

	c.renewal.RequestUpdate(certificate.CertificateMetaDomainJoiner{Certificate: a})
	return CertDbItem{
		Id:          a.Id,
		Owner:       a.Owner,
		AutoRenew:   utils.SBool(a.AutoRenew),
		Active:      utils.SBool(a.Active),
		Renewing:    utils.SBool(a.Renewing),
		RenewFailed: utils.SBool(a.RenewFailed),
		Domains:     d,
	}, err
}

func (c CertDbProvider) Get(id string) (CertDbItem, bool, error) {
	var z CertDbItem
	var a certificate.Certificate
	ok, err := c.db.Where("id = ?", id).Get(&a)
	if err == nil && ok {
		var b []certificate.CertificateDomain
		err := c.db.Where("cert_id = ?", a.Id).Find(&b)
		if err != nil {
			b = []certificate.CertificateDomain{}
		}
		d := make([]string, len(b))
		for i := range b {
			if utils.SBool(b[i].Wildcard) {
				d[i] = "*." + b[i].Domain
			} else {
				d[i] = b[i].Domain
			}
		}
		z = CertDbItem{
			Id:          a.Id,
			Owner:       a.Owner,
			AutoRenew:   utils.SBool(a.AutoRenew),
			Active:      utils.SBool(a.Active),
			Renewing:    utils.SBool(a.Renewing),
			RenewFailed: utils.SBool(a.RenewFailed),
			Domains:     d,
		}
	}
	return z, ok, err
}

func (c CertDbProvider) Put(id string, item CertDbItem) error {
	z := certificate.Certificate{AutoRenew: utils.PBool(item.AutoRenew), Active: utils.PBool(true)}
	_, err := c.db.Where("id = ?", id).Update(&z)
	return err
}

func (c CertDbProvider) Patch(id string, item CertDbItem) error {
	z := certificate.Certificate{AutoRenew: utils.PBool(item.AutoRenew), Active: utils.PBool(true)}
	_, err := c.db.Where("id = ?", id).Update(&z)
	return err
}

func (c CertDbProvider) Delete(id string) error {
	_, err := c.db.Where("id = ?", id).Update(&certificate.Certificate{Active: utils.PBool(false)})
	return err
}
