package main

import (
	quickUser "code.mrmelon54.com/melon/summer/cmd/marigold/quick-user"
	accountServer "code.mrmelon54.com/melon/summer/pkg/account-server"
	"code.mrmelon54.com/melon/summer/pkg/api"
	"code.mrmelon54.com/melon/summer/pkg/api/crud"
	"code.mrmelon54.com/melon/summer/pkg/cli"
	loginChecker "code.mrmelon54.com/melon/summer/pkg/login-checker"
	"code.mrmelon54.com/melon/summer/pkg/mailer"
	oauthServer "code.mrmelon54.com/melon/summer/pkg/oauth-server"
	"code.mrmelon54.com/melon/summer/pkg/tables/oauth"
	"code.mrmelon54.com/melon/summer/pkg/tables/user"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"crypto/rsa"
	"crypto/x509"
	"encoding/json"
	"encoding/pem"
	"fmt"
	"github.com/go-oauth2/oauth2/v4/manage"
	"github.com/go-oauth2/oauth2/v4/server"
	"github.com/gorilla/mux"
	"github.com/mrmelon54/mjwt"
	"math/rand"
	"net/http"
	"os"
	"strings"
	"time"
)

var usedTables = []any{
	&user.User{},
	&user.Perm{},
	&user.LinkUserPerm{},
	&oauth.AuthNamespace{},
	&oauth.Domain{},
	&oauth.App{},
}

func main() {
	cli.RunWithConfig[MarigoldConfig]("marigold", func(t MarigoldConfig) string { return t.Database }, new(Marigold))
}

type Marigold struct {
	conf         MarigoldConfig
	runner       *cli.Runner[MarigoldConfig]
	server       *http.Server
	signer       mjwt.Provider
	oauthManager *manage.Manager
	oauthSrv     *server.Server
	checkLogin   *loginChecker.LoginChecker
	auth         *accountServer.AccountServer
	oauth        *oauthServer.OAuthServer
	mailer       *mailer.Mailer
}

func (marigold *Marigold) Init(runner *cli.Runner[MarigoldConfig]) {
	marigold.runner = runner
	marigold.conf = runner.Config
	utils.Check("[Marigold.Init()] Failed to sync tables", runner.Database.Sync(usedTables...))
	quickUser.QuickUser(runner.Database)

	var privKey *rsa.PrivateKey
	file, err := os.ReadFile(marigold.conf.Auth.Key)
	if os.IsNotExist(err) {
		privKey = generateNewKeys(marigold.conf.Auth)
		file = pem.EncodeToMemory(&pem.Block{
			Type:  "RSA PRIVATE KEY",
			Bytes: x509.MarshalPKCS1PrivateKey(privKey),
		})
	} else {
		utils.Check("[Marigold.Init()] Failed to read token signing key", err)
		block, _ := pem.Decode(file)
		privKey, err = x509.ParsePKCS1PrivateKey(block.Bytes)
		utils.Check("[Marigold.Init()] Failed to parse token signing key", err)
	}
	marigold.signer = mjwt.NewMJwtSigner(marigold.conf.Auth.Issuer, privKey)
	marigold.checkLogin = loginChecker.NewLoginChecker(runner.Database, marigold.signer)

	marigold.mailer, _ = mailer.NewMailer(marigold.conf.Smtp)

	router := mux.NewRouter()
	marigold.auth = accountServer.NewAccountServer(router.PathPrefix("/auth").Subrouter(), runner.Database, marigold.signer, marigold.mailer)
	marigold.oauth = oauthServer.NewOAuthServer(router.PathPrefix("/oauth").Subrouter(), runner.Database, marigold.signer, marigold.conf.Auth.AuthClient)
	crud.NewCrudHandler[user.User](router, marigold.signer, accountServer.NewUserProvider(runner.Database), "/users", "/user/{id}")
	crud.NewCrudHandler[oauth.App](router, marigold.signer, oauthServer.NewOAuthAppProvider(runner.Database), "/apps", "/app/{id}")
	router.HandleFunc("/scopes", func(rw http.ResponseWriter, req *http.Request) {
		scopes := strings.Split(req.URL.Query().Get("scope"), " ")
		foundScopes := make([]user.Perm, 0)
		err = runner.Database.Where("is_scope = ?", true).In("key", scopes).Find(&foundScopes)
		if api.GenericErrorMsg[user.Perm](rw, err, http.StatusInternalServerError, "Failed to find scopes in database") {
			return
		}
		_ = json.NewEncoder(rw).Encode(foundScopes)
	})

	marigold.server = api.RunApiServer(marigold.conf.Listen, router)
}

func (marigold *Marigold) Destroy() {
	_ = marigold.server.Close()
}

func generateNewKeys(auth AuthConfig) *rsa.PrivateKey {
	// Generate key
	fmt.Println("[Marigold.Init()] Generating new RSA private key")
	key, err := rsa.GenerateKey(rand.New(rand.NewSource(time.Now().UnixNano())), 4096)
	utils.Check("[Marigold.Init()] Failed to generate new RSA private key", err)

	// Create key files
	createPriv, err := os.Create(auth.Key)
	utils.Check("[Marigold.Init()] Failed to open private key file for writing", err)
	createPub, err := os.Create(auth.Public)
	utils.Check("[Marigold.Init()] Failed to open public key file for writing", err)

	// Encode and write keys
	keyBytes := x509.MarshalPKCS1PrivateKey(key)
	pubBytes := x509.MarshalPKCS1PublicKey(&key.PublicKey)
	err = pem.Encode(createPriv, &pem.Block{Type: "RSA PRIVATE KEY", Bytes: keyBytes})
	utils.Check("[Marigold.Init()] Failed to encode private key to file", err)
	err = pem.Encode(createPub, &pem.Block{Type: "RSA PUBLIC KEY", Bytes: pubBytes})
	utils.Check("[Marigold.Init()] Failed to encode public key to file", err)
	return key
}
