package oauth_server

import (
	"code.mrmelon54.com/melon/summer/pkg/api"
	"code.mrmelon54.com/melon/summer/pkg/claims"
	"code.mrmelon54.com/melon/summer/pkg/claims/auth"
	"code.mrmelon54.com/melon/summer/pkg/claims/flow"
	"code.mrmelon54.com/melon/summer/pkg/tables/oauth"
	"code.mrmelon54.com/melon/summer/pkg/tables/user"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"encoding/json"
	"fmt"
	"github.com/go-oauth2/oauth2/v4/errors"
	"github.com/golang-jwt/jwt/v4"
	"github.com/google/uuid"
	"github.com/gorilla/mux"
	"github.com/mrmelon54/mjwt"
	"net/http"
	"net/url"
	"time"
	"xorm.io/xorm"
)

var ErrInvalidClient = errors.New("invalid client")

type OAuthServer struct {
	signer     mjwt.Provider
	router     *mux.Router
	db         *xorm.Engine
	authClient string
}

type OAuthFlowInput struct {
	ResponseType string              `json:"response_type"`
	ClientId     uint64              `json:"client_id"`
	Scope        *claims.PermStorage `json:"scope"`
	State        string              `json:"state"`
	RedirectUrl  string              `json:"redirect_url,omitempty"`
}

type OAuthFlowOutput struct {
	Code        string `json:"code"`
	RedirectUrl string `json:"redirect_url,omitempty"`
}

type OAuthTokenOutput struct {
	AccessToken  string              `json:"access_token"`
	TokenType    string              `json:"token_type"`
	ExpiresIn    uint64              `json:"expires_in,omitempty"`
	RefreshToken string              `json:"refresh_token,omitempty"`
	Scope        *claims.PermStorage `json:"scope,omitempty"`
	Sub          uint64              `json:"sub"`
	Aud          string              `json:"aud"`
	Exp          int64               `json:"exp"`
	IdToken      string              `json:"id_token"`
}

func NewOAuthServer(router *mux.Router, db *xorm.Engine, signer mjwt.Provider, authClient string) *OAuthServer {
	s := &OAuthServer{
		signer:     signer,
		router:     router,
		db:         db,
		authClient: authClient,
	}
	router.HandleFunc("/authorize", s.handleOAuthAuthorize).Methods(http.MethodPost, http.MethodGet)
	router.HandleFunc("/token", s.handleOAuthToken).Methods(http.MethodPost)
	router.HandleFunc("/info", s.handleOAuthInfo).Methods(http.MethodGet)
	return s
}

func ClientAuthHandler(req *http.Request) (string, string, error) {
	cId, cSecret, err := ClientBasicHandler(req)
	if cId == "" && cSecret == "" {
		cId, cSecret, err = ClientFormHandler(req)
	}
	return cId, cSecret, err
}

// ClientFormHandler get client data from form
func ClientFormHandler(r *http.Request) (string, string, error) {
	clientID := r.Form.Get("client_id")
	if clientID == "" {
		return "", "", ErrInvalidClient
	}
	clientSecret := r.Form.Get("client_secret")
	return clientID, clientSecret, nil
}

// ClientBasicHandler get client data from basic authorization
func ClientBasicHandler(r *http.Request) (string, string, error) {
	username, password, ok := r.BasicAuth()
	if !ok {
		return "", "", ErrInvalidClient
	}
	return username, password, nil
}

func (o *OAuthServer) handleOAuthAuthorize(rw http.ResponseWriter, req *http.Request) {
	if req.Method == http.MethodGet {
		appUrl, err := url.Parse(o.authClient)
		if err != nil {
			rw.WriteHeader(http.StatusInternalServerError)
			_, _ = rw.Write([]byte("Invalid application configuration\n"))
			return
		}
		appUrl.RawQuery = req.URL.RawQuery
		http.Redirect(rw, req, appUrl.String(), http.StatusFound)
		return
	}

	token := utils.GetBearerToken(req)
	if token == "" {
		api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, http.StatusInternalServerError, "Access token missing")
		return
	}

	var in OAuthFlowInput
	err := json.NewDecoder(req.Body).Decode(&in)
	if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Failed to decode JSON input") {
		return
	}

	// Verify as access token
	_, b, err := mjwt.ExtractClaims[auth.AccessTokenClaims](o.signer, token)
	if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusUnauthorized, "Token Not Valid") {
		return
	}

	if in.ResponseType == "code" {
		count, err := o.db.Where("app_id = ?", in.ClientId).Count(&oauth.App{})
		if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Invalid Client") {
			return
		}
		if count != 1 {
			api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, http.StatusBadRequest, "Invalid Client")
			return
		}

		redirect, err := url.Parse(in.RedirectUrl)
		if api.GenericErrorMsg[OAuthServer](rw, err, 419, "Invalid Redirect URL") {
			return
		}
		appUrl := fmt.Sprintf("%s://%s", redirect.Scheme, redirect.Host)
		n, err := o.db.Where("app_id = ? and domain = ?", in.ClientId, appUrl).Count(&oauth.Domain{})
		if api.GenericErrorMsg[OAuthServer](rw, err, 419, "Invalid Redirect URL") {
			return
		}
		if n < 1 {
			api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, 419, "Invalid Redirect URL")
			return
		}

		flowToken, err := flow.CreateOAuthFlowToken(o.signer, fmt.Sprint(in.ClientId), uuid.NewString(), b.Claims.UserId, in.Scope, time.Minute*10)
		if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Failed to sign token") {
			return
		}

		rw.Header().Set("Content-Type", "application/json")
		rw.Header().Set("X-Content-Type-Options", "nosniff")
		rw.WriteHeader(http.StatusAccepted)
		err = json.NewEncoder(rw).Encode(OAuthFlowOutput{Code: flowToken})
		api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Failed to encode JSON output")
	} else {
		api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, http.StatusBadRequest, "Invalid response type")
	}
}

func (o *OAuthServer) handleOAuthToken(rw http.ResponseWriter, req *http.Request) {
	err := req.ParseForm()
	if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Invalid POST request form") {
		return
	}

	cId, cSecret, err := ClientAuthHandler(req)
	if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusUnauthorized, "Unauthorized") {
		return
	}
	count, err := o.db.Where("app_id = ? and secret = ?", cId, cSecret).Count(&oauth.App{})
	if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Invalid Client") {
		return
	}
	if count != 1 {
		api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, http.StatusBadRequest, "Invalid Client")
		return
	}

	form := req.Form
	if form.Get("grant_type") == "authorization_code" {
		code := form.Get("code")

		// Verify as access token
		_, b, err := mjwt.ExtractClaims[flow.OAuthFlowClaims](o.signer, code)
		if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusUnauthorized, "Token Not Valid") {
			return
		}

		var u user.User
		ok, err := o.db.Where("id = ?", b.Claims.LoginId).Get(&u)
		if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Invalid User") {
			return
		}
		if !ok {
			api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Invalid User")
			return
		}

		token, err := auth.CreateAccessToken(o.signer, cId, fmt.Sprint(u.Id), u.Id, b.Claims.Scopes)
		if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusInternalServerError, "Token Signing Error") {
			return
		}

		rw.Header().Set("Content-Type", "application/json")
		rw.Header().Set("X-Content-Type-Options", "nosniff")
		rw.WriteHeader(http.StatusAccepted)
		n := time.Now()
		idToken, err := o.signer.SignJwt(jwt.RegisteredClaims{
			Issuer:    o.signer.Issuer(),
			Subject:   fmt.Sprint(u.Id),
			Audience:  jwt.ClaimStrings{cId},
			ExpiresAt: jwt.NewNumericDate(n.Add(time.Minute * 15)),
			NotBefore: jwt.NewNumericDate(n),
			IssuedAt:  jwt.NewNumericDate(n),
			ID:        fmt.Sprint(u.Id),
		})
		if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusInternalServerError, "ID Token Signing Error") {
			return
		}

		err = json.NewEncoder(rw).Encode(OAuthTokenOutput{
			AccessToken: token,
			TokenType:   "Bearer",
			ExpiresIn:   uint64((time.Minute * 15).Seconds()),
			Sub:         u.Id,
			Aud:         cId,
			Exp:         time.Now().Add(time.Minute * 15).Unix(),
			IdToken:     idToken,
		})
		api.GenericErrorMsg[OAuthServer](rw, err, http.StatusBadRequest, "Failed to encode JSON output")
	} else {
		api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, http.StatusBadRequest, "Invalid grant type")
	}
}

func (o *OAuthServer) handleOAuthInfo(rw http.ResponseWriter, req *http.Request) {
	token := utils.GetBearerToken(req)
	if token == "" {
		api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, http.StatusInternalServerError, "Access token missing")
		return
	}

	// Verify as access token
	_, b, err := mjwt.ExtractClaims[auth.AccessTokenClaims](o.signer, token)
	if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusUnauthorized, "Token Not Valid") {
		return
	}

	u := user.User{Id: b.Claims.UserId}
	get, err := o.db.Get(&u)
	if api.GenericErrorMsg[OAuthServer](rw, err, http.StatusInternalServerError, "Internal server error") {
		return
	}
	if !get {
		if api.GenericErrorMsg[OAuthServer](rw, api.ErrNoError, http.StatusForbidden, "Invalid JWT") {
			return
		}
	}

	a := map[string]interface{}{}
	a["sub"] = fmt.Sprint(u.Id)
	a["login"] = u.Username
	a["preferred_username"] = u.Username
	if b.Claims.Perms.Has("email") {
		a["email"] = u.Email
	}
	a["name"] = u.DisplayName
	a["given_name"] = u.DisplayName
	a["picture"] = u.Icon

	rw.Header().Set("Content-Type", "application/json")
	rw.WriteHeader(http.StatusOK)
	_ = json.NewEncoder(rw).Encode(a)
}
