package renewal

import (
	pebbleDev "code.mrmelon54.com/melon/summer/cmd/buttercup/pebble-dev"
	"code.mrmelon54.com/melon/summer/pkg/notifier"
	certificate2 "code.mrmelon54.com/melon/summer/pkg/tables/certificate"
	"code.mrmelon54.com/melon/summer/pkg/tables/renewal"
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"crypto"
	"crypto/tls"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"github.com/go-acme/lego/v4/certificate"
	"github.com/go-acme/lego/v4/lego"
	"github.com/go-acme/lego/v4/providers/dns/namesilo"
	"github.com/go-acme/lego/v4/registration"
	"github.com/pkg/errors"
	"io"
	"log"
	"net/http"
	"strings"
	"sync"
	"time"
	"xorm.io/xorm"
)

var ErrAlreadyRenewing = errors.New("renewal: already renewing")

type Renewal struct {
	db           *xorm.Engine
	httpProvider *CustomHTTPProvider
	certTicker   *time.Ticker
	certDone     chan struct{}
	caAddr       string
	caCert       []byte
	renewLock    *sync.Mutex
	notify       notifier.Notifier
}

type MyAccount struct {
	Email        string
	Registration *registration.Resource
	key          crypto.PrivateKey
}

func (u *MyAccount) GetEmail() string {
	return u.Email
}

func (u *MyAccount) GetRegistration() *registration.Resource {
	return u.Registration
}

func (u *MyAccount) GetPrivateKey() crypto.PrivateKey {
	return u.key
}

func NewRenewalService(wg *sync.WaitGroup, db *xorm.Engine, httpProvider *CustomHTTPProvider, config Config, notify notifier.Notifier) *Renewal {
	r := &Renewal{db: db, httpProvider: httpProvider, renewLock: &sync.Mutex{}, notify: notify}
	r.load(wg, config)
	return r
}

func (r *Renewal) logf(format string, v ...any) {
	s := strings.TrimSpace(fmt.Sprintf(format, v...))
	log.Println(s)
	r.notify.SendMessage(s)
}

func (r *Renewal) load(wg *sync.WaitGroup, config Config) {
	r.certTicker = time.NewTicker(time.Hour * 4)
	r.certDone = make(chan struct{})

	// Change CA dir for testing environment (for using pebble)
	r.caAddr = config.LetsEncrypt.Directory
	if config.LetsEncrypt.Certificate != "default" {
		if config.LetsEncrypt.Certificate == "pebble" {
			r.caCert = pebbleDev.GetPebbleCert()
		} else {
			// Download certificate
			caGet, err := http.Get(config.LetsEncrypt.Certificate)
			utils.Check("[Renewal] Failed to download CA certificate:", err)
			r.caCert, err = io.ReadAll(caGet.Body)
			utils.Check("[Renewal] Failed to read CA certificate:", err)
		}
	}

	wg.Add(1)
	go r.renewalRoutine(wg)
}

func (r *Renewal) setupLegoConfig(email string, privateKey crypto.PrivateKey) (*lego.Config, *MyAccount) {
	myUser := &MyAccount{Email: email, key: privateKey}
	config := lego.NewConfig(myUser)
	if r.caAddr != "production" {
		log.Printf("[Renewal] Overriding LEGO CADirUrl: '%s'\n", r.caAddr)
		if r.caAddr == "staging" {
			config.CADirURL = lego.LEDirectoryStaging
		} else {
			config.CADirURL = r.caAddr
		}
	}
	if r.caCert != nil {
		certPool := x509.NewCertPool()
		certPool.AppendCertsFromPEM(r.caCert)
		t := http.DefaultTransport.(*http.Transport).Clone()
		t.TLSClientConfig = &tls.Config{RootCAs: certPool}
		config.HTTPClient.Transport = t
	}
	return config, myUser
}

func (r *Renewal) renewalRoutine(wg *sync.WaitGroup) {
	defer func() {
		r.certTicker.Stop()
		log.Printf("[Renewal] Stopped certificate renewal service\n")
		wg.Done()
	}()

	log.Printf("[Renewal] Doing quick certificate check before starting...\n")
	err := r.renewalCheck()
	if err != nil {

	}
	log.Printf("[Renewal] Initial check complete, continually checking every 4 hours...\n")

	for {
		select {
		case <-r.certDone:
			return
		case <-r.certTicker.C:
			_ = r.renewalCheck()
		}
	}
}

func (r *Renewal) renewalCheck() error {
	if !r.renewLock.TryLock() {
		return ErrAlreadyRenewing
	}
	r.renewLock.Unlock()

	var certs []certificate2.CertificateMetaDomainJoiner
	//goland:noinspection SqlDialectInspection,SqlNoDataSourceInspection
	err := r.db.SQL(`
select *
from certificate_data
         inner join certificate on certificate.id = certificate_data.meta_id
where certificate.active = 1
  and certificate.auto_renew = 1
  and certificate.renewing = 0
  and certificate.renew_failed = 0
  and certificate_data.ready = 1
  and DATE_ADD(certificate_data.not_after, INTERVAL -30 DAY) < now()
order by not_after DESC
`).Find(&certs)
	if err != nil {
		r.logf("[Renewal::renewalCheck()] Failed to get certificates for a renewal check: %s", err)
		return nil
	}

	switch {
	case len(certs) == 1:
		r.logf("[Renewal] Cert '%d - %d' runs out in 30 days", certs[0].Id, certs[0].DataId)
		r.RequestUpdate(certs[0])
	case len(certs) > 1:
		r.logf("[Renewal] Cert '%d - %d' runs out in 30 days, so do %d other certs which will update one-by-one each 4 hours", certs[0].Id, certs[0].DataId, len(certs)-1)
		r.RequestUpdate(certs[0])
	}

	return nil
}

func (r *Renewal) RequestUpdate(c certificate2.CertificateMetaDomainJoiner) {
	var a []certificate2.CertificateData
	err := r.db.Where("previous_cert = ?", c.DataId).Limit(1).Find(&a)
	if err != nil {
		r.logf("[Renewal::RequestUpdate()] Failed to update certificate: %s", err)
		return
	}
	if len(a) >= 1 {
		r.logf("[Renewal] Unable to update certificate '%d'... Found a newer certificate '%d'", c.Id, a[0].DataId)
		return
	}

	if c.Renewing == nil {
		c.Renewing = utils.PBool(false)
	}
	if !*c.AutoRenew {
		log.Printf("[Renewal] Unable to update certificate '%d'... Auto renewal is disabled\n", c.Id)
		return
	}
	if *c.Renewing {
		log.Printf("[Renewal] Unable to update certificate '%d'... It is already renewing\n", c.Id)
		return
	}

	log.Printf("[Renewal] Requested force update for certificate '%d'\n", c.Id)
	go r.updateCertificate(&c)
}

func (r *Renewal) Shutdown() {
	log.Printf("[Renewal] Shutting down certificate renewal service")
	close(r.certDone)
}

func (r *Renewal) fetchDomains(c *certificate2.Certificate) []string {
	var domains []certificate2.CertificateDomain
	err := r.db.Where("cert_id = ?", c.Id).Find(&domains)
	if err != nil {
		log.Println("[fetchDomains()] Error:", err)
		return []string{}
	}

	l := len(domains)
	allDomains := make([]string, l)
	for k, v := range domains {
		if v.Wildcard != nil && *v.Wildcard {
			allDomains[k] = "*." + v.Domain
		} else {
			allDomains[k] = v.Domain
		}
	}

	if len(allDomains) == 0 {
		log.Printf("[Renewal] All domains for certificate '%d' is empty. (%s)\n", c.Id, strings.Join(allDomains, ", "))
	}

	return allDomains
}

func (r *Renewal) updateCertificate(c *certificate2.CertificateMetaDomainJoiner) {
	domains := r.fetchDomains(&c.Certificate)
	r.logf("[Renewal] Starting renewal service request for '%s' (%d)", strings.Join(domains, ", "), c.Id)

	// switch to renewing state
	c.Renewing = utils.PBool(true)
	_, err := r.db.Where("id = ?", c.Id).Update(&c.Certificate)
	if err != nil {
		r.logf("[Renewal] Failed to update certificate state to (renewing): %s", err)
		return
	}

	// setup LetsEncrypt client
	client, err := r.setupForRequest(c)
	if err != nil {
		r.logf("[Renewal] Failed to generate a client for requesting a certificate: %s", err)
		return
	}

	// use LetsEncrypt to obtain a new certificate
	request := certificate.ObtainRequest{
		Domains: domains,
		Bundle:  true,
	}
	certificates, err := client.Certificate.Obtain(request)
	if err != nil {
		r.logf("[Renewal] Failed to obtain replacement certificate for '%d': %s", c.Id, err)
		return
	}

	// create new certificate data
	newCert := certificate2.CertificateData{
		MetaId:           c.MetaId,
		PreviousCert:     c.DataId,
		CertificateBytes: certificates.Certificate,
		KeyBytes:         certificates.PrivateKey,
		Ready:            utils.PBool(true),
	}
	err = newCert.UpdateNotAfter()
	if err != nil {
		r.logf("[Renewal] Failed to update NotAfter value: %s", err)
		newCert.Ready = utils.PBool(false)
		_, _ = r.db.Insert(&newCert)
		c.Renewing = utils.PBool(false)
		c.RenewFailed = utils.PBool(true)
		_, err = r.db.Where("id = ?", c.Id).Update(&c.Certificate)
		if err != nil {
			r.logf("[Renewal] Failed to update state to (not renewing, renew failed): %s", err)
		}
		return
	}

	// insert new certificate
	newCert.PreviousCert = c.DataId
	_, err = r.db.Insert(&newCert)
	if err != nil {
		r.logf("[Renewal] Failed to insert new certificate: %s", err)
		return
	}

	cd2 := certificate2.CertificateData{Ready: utils.PBool(false)}
	_, err = r.db.Update(&cd2, &certificate2.CertificateData{DataId: c.DataId})
	if err != nil {
		r.logf("[Renewal] Failed to disable previous certificate: %s", err)
		return
	}

	// update meta to usable state
	c.Renewing = utils.PBool(false)
	c.RenewFailed = utils.PBool(false)
	_, err = r.db.Where("id = ?", c.Id).Update(&c.Certificate)
	if err != nil {
		r.logf("[Renewal] Failed to update state to (not renewing, not renew failed): %s", err)
		return
	}

	r.logf("[Renewal] Successfully generated replacement certificate '%d' for '%d'", newCert.DataId, c.Id)
	return
}

func (r *Renewal) setupForRequest(c *certificate2.CertificateMetaDomainJoiner) (*lego.Client, error) {
	var letsEncryptAccounts []renewal.LetsEncryptAccount
	_ = r.db.Limit(1).Find(&letsEncryptAccounts, renewal.LetsEncryptAccount{Id: c.LetsEncrypt})
	if len(letsEncryptAccounts) != 1 {
		return nil, errors.Errorf("[Renewal] Couldn't load Lets Encrypt account '%d'", c.LetsEncrypt)
	}

	var namesiloAccount renewal.NamesiloAccount
	usingNamesilo, _ := r.db.Limit(1).Where("id = ?", c.Namesilo).Get(&namesiloAccount)

	myAccount := letsEncryptAccounts[0]

	letsEncryptPemBlock, _ := pem.Decode(myAccount.PrivateKey)
	letsEncryptPrivKey, err := x509.ParsePKCS1PrivateKey(letsEncryptPemBlock.Bytes)
	if err != nil {
		return nil, errors.Errorf("[Renewal] Failed to parse Lets Encrypt private key '%d': %s", myAccount.Id, err)
	}

	legoConfig, myUser := r.setupLegoConfig(myAccount.Email, letsEncryptPrivKey)
	client, err := lego.NewClient(legoConfig)
	if err != nil {
		return nil, errors.Errorf("[Renewal] Couldn't load LEGO client: %s", err)
	}

	if usingNamesilo {
		log.Printf("Using namesilo account: %s - %s%s\n", namesiloAccount.Email, namesiloAccount.Token[:5], strings.Repeat("*", 16))
		dnsNamesilo, err := namesilo.NewDNSProviderConfig(&namesilo.Config{
			APIKey:             namesiloAccount.Token,
			PropagationTimeout: 2 * time.Hour,
			PollingInterval:    15 * time.Minute,
			TTL:                3600,
		})
		if err != nil {
			return nil, errors.Errorf("[Renewal] Failed to setup DNS provider '%d'\n", namesiloAccount.Id)
		}
		err = client.Challenge.SetDNS01Provider(dnsNamesilo)
		if err != nil {
			return nil, errors.Errorf("[Renewal] Failed to attach DNS provider '%d'\n", namesiloAccount.Id)
		}
	}

	err = client.Challenge.SetHTTP01Provider(r.httpProvider)
	if err != nil {
		return nil, errors.Errorf("[Renewal] Failed to attach HTTP provider")
	}

	qReg, err := client.Registration.Register(registration.RegisterOptions{TermsOfServiceAgreed: true})
	if err != nil {
		return nil, errors.Errorf("[Renewal] Failed to update account registration: %s", err)
	}
	myUser.Registration = qReg
	return client, nil
}

func (r *Renewal) Renew() {}
