package udp

import (
	"code.mrmelon54.com/melon/summer/pkg/tables/web"
	"fmt"
	"github.com/gorilla/mux"
	"log"
	"net"
	"sync"
	"xorm.io/xorm"
)

type Manager struct {
	mutex   *sync.RWMutex
	servers map[uint64]*Proxy
	router  *mux.Router
	db      *xorm.Engine
}

func NewUdpManager(db *xorm.Engine) *Manager {
	m := &Manager{
		mutex:   &sync.RWMutex{},
		servers: make(map[uint64]*Proxy),
		db:      db,
	}
	m.Compile()
	return m
}

func (m *Manager) Compile() {
	m.empty()
	var a []web.UdpRedirect
	err := m.db.Where("enabled = ?", true).Find(&a)
	if err != nil {
		log.Println("[UdpManager] Failed to open startup UDP ports")
		a = []web.UdpRedirect{}
	}
	for _, i := range a {
		m.Add(i)
	}
}

func (m *Manager) Open(n uint64, lAddr, rAddr *net.UDPAddr) error {
	log.Printf("[UdpManager] Opening %d: %s => %s\n", n, lAddr, rAddr)
	p, err := NewUdpProxy(lAddr, rAddr)
	if err != nil {
		return err
	}
	m.mutex.Lock()
	m.servers[n] = p
	m.mutex.Unlock()
	return nil
}

func (m *Manager) Close(n uint64) {
	m.mutex.Lock()
	if p, ok := m.servers[n]; ok {
		log.Printf("[UdpManager] Closing %d (%s => %s)\n", n, p.lAddr, p.rAddr)
		p.Quit()
		delete(m.servers, n)
	}
	m.mutex.Unlock()
}

func (m *Manager) empty() {
	wg := &sync.WaitGroup{}
	m.mutex.Lock()
	for _, p := range m.servers {
		wg.Add(1)
		go func(p *Proxy) {
			defer wg.Done()
			p.Quit()
		}(p)
	}
	wg.Wait()
	m.servers = make(map[uint64]*Proxy)
	m.mutex.Unlock()
}

func (m *Manager) Destroy() {
	log.Println("[UdpManager] Destroying UDP Manager")
	m.empty()
}

func (m *Manager) Add(item web.UdpRedirect) {
	lAddr, err := net.ResolveUDPAddr("udp", fmt.Sprintf(":%d", item.Port))
	if err != nil {
		return
	}
	rAddr, err := net.ResolveUDPAddr("udp", item.Target)
	if err != nil {
		return
	}
	_ = m.Open(item.Id, lAddr, rAddr)
}
