package udp

import (
	"code.mrmelon54.com/melon/summer/pkg/utils"
	"log"
	"net"
	"sync"
)

type Proxy struct {
	lAddr *net.UDPAddr
	rAddr *net.UDPAddr
	done  *utils.DoneChan
	l     *net.UDPConn
	n     uint64
	cMux  *sync.RWMutex
	cMap  map[string]*SingleConn
}

type SingleConn struct {
	l *net.UDPAddr
	r *net.UDPConn
}

func NewUdpProxy(lAddr, rAddr *net.UDPAddr) (*Proxy, error) {
	p := &Proxy{
		lAddr: lAddr,
		rAddr: rAddr,
		done:  utils.NewDoneChan(),
		cMux:  &sync.RWMutex{},
		cMap:  make(map[string]*SingleConn),
	}
	l, err := net.ListenUDP("udp", p.lAddr)
	if err != nil {
		return nil, err
	}
	p.l = l
	go p.listen()
	return p, err
}

func (p *Proxy) Quit() {
	p.done.Close()
	_ = p.l.Close()
}

func (p *Proxy) listen() {
	buf := make([]byte, 1500)
outer:
	for {
		// read data
		i, addr, err := p.l.ReadFromUDP(buf)
		if err != nil {
			log.Printf("[WARN %s @ %s] Read failed '%s'\n", addr, p.lAddr, err)
			continue outer
		}
		data := buf[:i]

		p.cMux.RLock()
		c, ok := p.cMap[addr.String()]
		p.cMux.RUnlock()
		if !ok {
			c = p.setupConn(addr)
			go p.runConn(c)
		}

		// relay to server
		_, err = c.r.Write(data)
		if err != nil {
			log.Printf("[WARN %s @ %s] Write failed '%s'\n", addr, p.lAddr, err)
		}
	}
}

func (p *Proxy) setupConn(addr *net.UDPAddr) *SingleConn {
	r, err := net.DialUDP("udp", nil, p.rAddr)
	if err != nil {
		log.Printf("[WARN %s @ %s] Dial UDP failed '%s'\n", addr, p.lAddr, err)
		return nil
	}
	c := &SingleConn{
		l: addr,
		r: r,
	}
	p.cMux.Lock()
	p.cMap[addr.String()] = c
	p.cMux.Unlock()
	return c
}

func (p *Proxy) runConn(conn *SingleConn) {
	buf := make([]byte, 1500)
	for {
		// read data
		i, err := conn.r.Read(buf)
		if err != nil {
			log.Printf("[WARN %s @ %s] Read failed '%s'\n", p.lAddr, p.rAddr, err)
			break
		}
		data := buf[:i]

		// write data
		_, err = p.l.WriteToUDP(data, conn.l)
		if err != nil {
			log.Printf("[WARN %s @ %s] Write failed '%s'\n", p.lAddr, p.rAddr, err)
			break
		}
	}
	p.cMux.Lock()
	delete(p.cMap, conn.l.String())
	p.cMux.Unlock()
}
